/* 
 * File:   BuildEstimator.cpp
 * Author: chris
 * 
 * Created on April 21, 2014, 4:35 PM
 */

#include "BuildEstimator.h"

BuildEstimator::~BuildEstimator()
{
}

UnitCommandRequest BuildEstimator::MakeNewCommandRequest(const BWAPI::UnitType& type, const int& unitTask)
{
    return UnitCommandRequest(NextUnitId, type, unitTask);
}

const int BuildEstimator::AddNewPlayer()
{
    const int playerId = NextPlayerId++;
    PlayerBuildState newPlayerState;
    for (int i = 0; i < newPlayerState.GetGasSCVCount(); i++, NextUnitId++)
    {
        newPlayerState.AddDefaultUnit(MakeNewCommandRequest(BWAPI::UnitTypes::Terran_SCV, 1));
    }

    for (int i = 0; i < newPlayerState.GetMineralSCVcount(); i++, NextUnitId++)
    {
        newPlayerState.AddDefaultUnit(MakeNewCommandRequest(BWAPI::UnitTypes::Terran_SCV, 0));
    }

    for (int i = 0; i < newPlayerState.GetRefineryCount(); i++, NextUnitId++)
    {
        newPlayerState.AddDefaultUnit(MakeNewCommandRequest(BWAPI::UnitTypes::Terran_Refinery, 0));
    }

    for (int i = 0; i < newPlayerState.GetBaseCount(); i++, NextUnitId++)
    {
        newPlayerState.AddDefaultUnit(MakeNewCommandRequest(BWAPI::UnitTypes::Terran_Command_Center, 0));
    }
    playerStates.insert(pair<int,PlayerBuildState>(playerId,newPlayerState));
    return playerId;
}

const QueueResult BuildEstimator::QueueCommand(const int& playerId, const BWAPI::UnitType& type, const int& unitTask)
{
    UnitCommandRequest newReq = MakeNewCommandRequest(type, unitTask);
    
    if (playerStates[playerId].QueueCommand(newReq, currentTimeInFrames))
    {
        NextUnitId++;
        return Success;
    }
    else if(playerStates[playerId].NeverSuccessfullyQueued(newReq))
    {
        return Deadlock;
    }        
    return Stall;
}

const int BuildEstimator::GetPlayerCount() const
{
    return playerStates.size();
}

const int BuildEstimator::GetCurrentFrame()
{
    return currentTimeInFrames;
}

const bool BuildEstimator::IncreaseCurrentFrame(const int& frameCount)
{
    currentTimeInFrames += frameCount;
    for (map<int, PlayerBuildState>::iterator player = playerStates.begin(); player != playerStates.end(); player++)
    {
        player->second.IncreaseCurrentFrame(frameCount, currentTimeInFrames);
    }
    return true;
}

const list<UnitCommandQueued>& BuildEstimator::GetCompleteCommands(const int& player)
{
    return playerStates.at(player).GetCompleteCommands();
}

const int BuildEstimator::GetCompletedUnitCount(const int& player, const BWAPI::UnitType& type)
{
    int unitCount = 0;
    list<UnitCommandQueued> commands = GetCompleteCommands(player);
    for(list<UnitCommandQueued>::const_iterator c = commands.begin(); c != commands.end(); c++)
    {
        if(c->type == type)
        {
            unitCount++;
        }
    }
    return unitCount;
}

//const bool BuildEstimator::CreateStartingUnits(const UnitCommandRequest& command, PlayerBuildState& player)
//{
//    
//}