/* 
 * File:   PlayerBuildState.cpp
 * Author: chris
 * 
 * Created on April 21, 2014, 4:35 PM
 */

#include "PlayerBuildState.h"
#include "UnitCommandStructs.h"
#include <float.h>
#include <set>
#include <climits>
#include "DEBUG.h"
map<BWAPI::UnitType, UnitSupplBuildStats> PlayerBuildState::buildInfo;

void PlayerBuildState::initSupplInfo()
{
    //    for (set<BWAPI::UnitType> unitType = BWAPI::unitTypeSet.begin(); unitType != BWAPI::unitTypeSet.end(); unitType++)
    //    {
    //        buildInfo[*unitType] = UnitSupplBuildStats(); // init all to zero
    //    }
    //give builder units correct value
    buildInfo[BWAPI::UnitTypes::Terran_SCV] = UnitSupplBuildStats(1);
    buildInfo[BWAPI::UnitTypes::Terran_Command_Center] = UnitSupplBuildStats(1);
    buildInfo[BWAPI::UnitTypes::Terran_Factory] = UnitSupplBuildStats(1);
    buildInfo[BWAPI::UnitTypes::Terran_Barracks] = UnitSupplBuildStats(1);
}

bool SortCompletionTimeLowToHigh(UnitCommandQueued& t1, UnitCommandQueued& t2) // Returns true if t1 is earlier than t2
{
    if (t1.frameCompleted < t2.frameCompleted)
        return true;
    return false;
}

bool SortCompletionTimeHighToLow(UnitCommandQueued& t1, UnitCommandQueued& t2) // Returns true if t1 is earlier than t2
{
    if (t1.frameCompleted < t2.frameCompleted)
        return false;
    return true;
}

//PlayerBuildState::PlayerBuildState()
//{
//}
//
//PlayerBuildState::PlayerBuildState(const PlayerBuildState& orig)
//{
//}

PlayerBuildState::~PlayerBuildState()
{
}

const bool PlayerBuildState::QueueCommand(const UnitCommandRequest& command, const int& currentTime)
{
    if (CheckForCompletePrereqStructs(command) && CheckForAvailableBuildQueue(command) && CheckForRequiredResourcesCost(command))
    {
        pair<int, double> soonestQueue = FindEarliestBuilderQueue(command.type);
        soonestQueue.second = (soonestQueue.second < currentTime ? currentTime : soonestQueue.second); //make sure the soonest start time is at least the current time
        double completedFrame = soonestQueue.second + command.type.buildTime();
        if(command.type.whatBuilds().first == BWAPI::UnitTypes::Terran_SCV)
        {
            completedFrame += scvBuildDelay;
        }
        UnitCommandQueued newQueuedCommand(command.unitId, soonestQueue.first, completedFrame, false, command.type, command.unitTask);
        incompleteCommands.push_back(newQueuedCommand);
        incompleteUnitTypes.insert(pair<BWAPI::UnitType, UnitCommandQueued>(newQueuedCommand.type, newQueuedCommand));
        buildQueueTracker.insert(pair<int, pair<int, double> >(newQueuedCommand.builderId, pair<int, double>(newQueuedCommand.unitId, newQueuedCommand.frameCompleted)));
        if (newQueuedCommand.type.whatBuilds().first == BWAPI::UnitTypes::Terran_SCV)
        {
            availbleSCVTracker.erase(newQueuedCommand.builderId);
            buildingSCVTracker.insert(pair<int, int>(newQueuedCommand.builderId, newQueuedCommand.unitId));
            mineralSCVcount--;
        }
        currentMinerals -= newQueuedCommand.type.mineralPrice();
        currentGas -= newQueuedCommand.type.gasPrice();
        DOUT("COMMAND QUEUED: " << newQueuedCommand.unitId << " In queue " << newQueuedCommand.builderId << " | " << newQueuedCommand.frameCompleted <<  "| Resources" << currentMinerals << "|"<<currentGas << endl;);
        return true;
    }
    return false;
}

void PlayerBuildState::UpdateUnits(const int& currentFrame)
{
    list<UnitCommandQueued> newlyCompletedUnits;
    //compete newly completed commands and remove from incomplete list
    incompleteCommands.sort(SortCompletionTimeLowToHigh);
    while (incompleteCommands.size() > 0 && incompleteCommands.front().frameCompleted <= currentFrame)
    {
        newlyCompletedUnits.push_back(incompleteCommands.front());
        incompleteCommands.pop_front();
    }

    //updates for each new unit
    //DOUT("Newly completed units: " << newlyCompletedUnits.size() << endl);
    while (newlyCompletedUnits.size() > 0)
    {
        //update new resource unit count
        AddCompletedUnit(newlyCompletedUnits.front());
        if (newlyCompletedUnits.front().type.whatBuilds().first == BWAPI::UnitTypes::Terran_SCV) //if we completed something that is built by an SCV, we can return that SCV to gathering minerals
        {
            buildingSCVTracker.erase(newlyCompletedUnits.front().builderId);
            availbleSCVTracker.insert(newlyCompletedUnits.front().builderId);
            mineralSCVcount++;
        }

        //update Build tracker build queue by removing completed unit
        for (map<int, pair<int, int> >::iterator built = buildQueueTracker.lower_bound(newlyCompletedUnits.front().builderId); built != buildQueueTracker.upper_bound(newlyCompletedUnits.front().builderId); built++)
        {
            if (built->second.first == newlyCompletedUnits.front().unitId)
            {
                buildQueueTracker.erase(built);
                break;
            }
        }

        //update completed commands by type
        for (multimap<BWAPI::UnitType, UnitCommandQueued>::iterator built = incompleteUnitTypes.lower_bound(newlyCompletedUnits.front().type); built != incompleteUnitTypes.upper_bound(newlyCompletedUnits.front().type); built++)
        {
            if (built->second.unitId == newlyCompletedUnits.front().unitId)
            {
                incompleteUnitTypes.erase(built);
                //completeUnitTypes.insert(pair<BWAPI::UnitType, UnitCommandQueued>(newlyCompletedUnits.front().type, newlyCompletedUnits.front()));
                break;
            }
        }
        newlyCompletedUnits.pop_front();
    }
    //update completed commands
    //completeCommands.merge(newlyCompletedUnits, SortCompletionTimeLowToHigh);
}

double PlayerBuildState::ResourcesPerFrame(const int& unitCount, const double& unitRPS, const double& scvCapPerNode, const double& nodes) const
{
    const float& unitCap = scvCapPerNode*nodes;
    float activeUnits = (unitCount > unitCap ? unitCap : unitCount);
    return activeUnits*unitRPS;
}

void PlayerBuildState::UpdateResources(const int& frameCount)
{
    currentMinerals += frameCount * ResourcesPerFrame(mineralSCVcount, mineralsPerFramePerSCV, scvCapPerMineralNode, avgMineralNodesPerBase * baseCount);
    currentGas += frameCount * ResourcesPerFrame(gasSCVcount, gasPerFramePerSCV, scvCapPerGasNode, refineryCount);
}

const bool PlayerBuildState::IncreaseCurrentFrame(const int& frameCount, const int& currentFrame)
{
    UpdateResources(frameCount);
    UpdateUnits(currentFrame);
}

const int PlayerBuildState::UnitTypeCompleteCount(const BWAPI::UnitType& unit, const int& task = 0)
{
    if (unit == BWAPI::UnitTypes::Terran_SCV && task == 0)
    {
        //DOUT("Update available SCV: " << availbleSCVTracker.size() << "|" << buildingSCVTracker.size() << endl);
        return availbleSCVTracker.size() + buildingSCVTracker.size();
    }
    else if (unit == BWAPI::UnitTypes::Terran_SCV && task == 1)
    {
        return completeUnitTypes.count(unit) - availbleSCVTracker.size() - buildingSCVTracker.size(); // all scvs - mineral scvs = gas scv count
    }
    else
    {
        return completeUnitTypes.count(unit);
    }
    //return (completeUnitTypes.find(unit) == completeUnitTypes.end() ? false : true);
    //    if (completeUnitTypes.find(unit) == completeUnitTypes.end())
    //    {
    //        return false;
    //    }
    //    return true;
}

const int PlayerBuildState::UnitTypeIncompleteCount(const BWAPI::UnitType& unit, const int& task = 0)
{
    if (unit == BWAPI::UnitTypes::Terran_SCV && task == 0)
    {
        int scvCount = 0;
        for (multimap<BWAPI::UnitType, UnitCommandQueued>::const_iterator scv = incompleteUnitTypes.lower_bound(unit); scv != incompleteUnitTypes.upper_bound(unit); scv++)
        {
            if (scv->second.unitTask == 0)
            {
                scvCount++;
            }

        }
        return scvCount; //availbleSCVTracker.size() + buildingSCVTracker.size();
    }
    else if (unit == BWAPI::UnitTypes::Terran_SCV && task == 1)
    {
        int scvCount = 0;
        for (multimap<BWAPI::UnitType, UnitCommandQueued>::const_iterator scv = incompleteUnitTypes.lower_bound(unit); scv != incompleteUnitTypes.upper_bound(unit); scv++)
        {
            if (scv->second.unitTask == 1)
            {
                scvCount++;
            }

        }
        return scvCount; //availbleSCVTracker.size() + buildingSCVTracker.size();
    }
    else
    {
        return incompleteUnitTypes.count(unit);
    }
    //return (incompleteUnitTypes.find(unit) == incompleteUnitTypes.end() ? false : true);
}

const bool PlayerBuildState::CheckForCompletePrereqStructs(const UnitCommandRequest& command)
{
    map<BWAPI::UnitType, int> prereqs = command.type.requiredUnits();
    for (map<BWAPI::UnitType, int>::const_iterator p = prereqs.begin(); p != prereqs.end(); p++)
    {
        if (UnitTypeCompleteCount(p->first, 0) == 0)
        {
            return false;
        }
    }
    //check for constructing unit
    if (UnitTypeCompleteCount(command.type.whatBuilds().first, 0) == 0)
    {
        return false;
    }
    return true;
}

const bool PlayerBuildState::CheckForAvailableBuildQueue(const UnitCommandRequest& command)
{
    BWAPI::UnitType builder = command.type.whatBuilds().first;
    //list of completed builder units for completed command
    if (builder == BWAPI::UnitTypes::Terran_SCV)
    {
        if (availbleSCVTracker.size() > 0)
        {
            return true;
        }
    }
    else
    {
        for (multimap<BWAPI::UnitType, UnitCommandQueued>::const_iterator bunit = completeUnitTypes.lower_bound(builder); bunit != completeUnitTypes.upper_bound(builder); bunit++)
        {
            //empty queue, stop searching
            if (buildQueueTracker.lower_bound(bunit->second.unitId) == buildQueueTracker.upper_bound(bunit->second.unitId))
            {
                return true;
            }
            else if (buildInfo[bunit->first].MAX_BUILD_SLOTS > buildQueueTracker.count(bunit->second.unitId)) //queue isnt full, return true
            {
                return true;
            }
        }
    }
    return false;
}

const bool PlayerBuildState::CheckForRequiredResourcesCost(const UnitCommandRequest& command)
{
    //check for enough resources to build
    //DOUT("Cost to build unit: " << command.type.mineralPrice() << " | " << command.type.gasPrice() << endl);
    if (command.type.gasPrice() > currentGas || command.type.mineralPrice() > currentMinerals)
    {
        return false;
    }
    return true;
}

const bool PlayerBuildState::CheckForRequiredResourcesGathering(const UnitCommandRequest& command)
{
    //check not gathering right resources
    //DOUT("Cost to build unit: " << command.type.mineralPrice() << " | " << command.type.gasPrice() << endl);
    if (command.type.gasPrice() > currentGas && 
            (UnitTypeCompleteCount(BWAPI::UnitTypes::Terran_Refinery) == 0 && UnitTypeIncompleteCount(BWAPI::UnitTypes::Terran_Refinery) == 0 || 
            UnitTypeCompleteCount(BWAPI::UnitTypes::Terran_SCV,1) == 0 && UnitTypeIncompleteCount(BWAPI::UnitTypes::Terran_SCV,1) == 0))
    {
        return false;
    }
    else if (command.type.mineralPrice() > currentMinerals && UnitTypeCompleteCount(BWAPI::UnitTypes::Terran_SCV,0) == 0 && UnitTypeIncompleteCount(BWAPI::UnitTypes::Terran_SCV,0) == 0)
    {
        return false;
    }
    return true;
}

const bool PlayerBuildState::NeverSuccessfullyQueued(const UnitCommandRequest& command)
{
    map<BWAPI::UnitType, int> prereqs = command.type.requiredUnits();
    for (map<BWAPI::UnitType, int>::const_iterator p = prereqs.begin(); p != prereqs.end(); p++)
    {
        //havent built and never plan to build a prereq unit
        if (UnitTypeCompleteCount(p->first) == 0 && UnitTypeIncompleteCount(p->first) == 0)
        {
            return true;
        }
    }
    if (!CheckForRequiredResourcesGathering(command))
    {
        return true;
    }

    return false;
}

pair<int, int> PlayerBuildState::FindEarliestBuilderQueue(const BWAPI::UnitType& unitToBuild) //return is builder ID and lastUnitOut frame
{
    BWAPI::UnitType builder = unitToBuild.whatBuilds().first;
    pair<int, int> soonestQueue = pair<int, int>(-1, INT_MAX);
    if (builder == BWAPI::UnitTypes::Terran_SCV)
    {
        //any SCV that is available can always start immediately (only 1 build slot)
        if (availbleSCVTracker.size() > 0)
        {
            soonestQueue.first = (*availbleSCVTracker.begin());
            soonestQueue.second = -INT_MAX;
        }
    }
    else
    {
        //multimap<BWAPI::UnitType,int> completeUnitTypes;
        //check each completed unit for the builder type, stop looping early on -DBL_MAX (command can start instantly, no need to check for better result)
        for (multimap<BWAPI::UnitType, UnitCommandQueued>::const_iterator bunit = completeUnitTypes.lower_bound(builder); bunit != completeUnitTypes.upper_bound(builder) && soonestQueue.second > -DBL_MAX; bunit++)
        {
            pair<int, int> latestQueue = pair<int, int>(-1, -INT_MAX);

            //find the farthest completion time for each individual builder unit, and take the builder unit with the smallest value
            //check to make sure the builder units queue isnt completely open(queue size 0 / does not exist, 0 wait time)
            if (buildQueueTracker.lower_bound(bunit->second.unitId) == buildQueueTracker.upper_bound(bunit->second.unitId))
            {
                latestQueue.first = bunit->second.unitId;
                latestQueue.second = -INT_MAX; //keep latest time as -MAX_DBL (queued command starts instantly)
                // latestQueue.second = bqueue->second->second; //keep latest time as -MAX_DBL (queued command starts instantly)
            }
            else if (buildInfo[bunit->first].MAX_BUILD_SLOTS > buildQueueTracker.count(bunit->second.unitId)) //only check queue for soonest time if the queue isnt full (if its full, we cant insert into it)
            {
                for (multimap<int, pair<int, int> >::iterator bqueue = buildQueueTracker.lower_bound(bunit->second.unitId); bqueue != buildQueueTracker.upper_bound(bunit->second.unitId); bqueue++)
                {
                    if (bqueue->second.second > latestQueue.second)
                    {
                        latestQueue.first = bqueue->first;
                        latestQueue.second = bqueue->second.second;
                    }
                }
            }
            else //max out the queue time, since it's unusable/full
            {
                latestQueue.second = INT_MAX;
            }
            if (latestQueue.second < soonestQueue.second)
            {
                soonestQueue = latestQueue;
            }
        }
    }
    return soonestQueue;

}

const double PlayerBuildState::GetCurrentGas()
{
    return currentGas;
}

const double PlayerBuildState::GetCurrentMinerals()
{
    return currentMinerals;
}

const double PlayerBuildState::GetMineralSCVcount()
{
    return mineralSCVcount;
}

const double PlayerBuildState::GetGasSCVCount()
{
    return gasSCVcount;
}

const double PlayerBuildState::GetBaseCount()
{
    return baseCount;
}

const double PlayerBuildState::GetRefineryCount()
{
    return refineryCount;

}

const bool PlayerBuildState::AddDefaultUnit(const UnitCommandRequest& command)
{
    UnitCommandQueued newCommand(command.unitId, -1, 0, true, command.type, command.unitTask);
    //AddCompletedUnit(newCommand);
    if (command.type == BWAPI::UnitTypes::Terran_SCV && command.unitTask == 0)
    {
        availbleSCVTracker.insert(command.unitId);
        //DOUT("Update available SCV: " << availbleSCVTracker.size() << endl);
    }

    completeUnitTypes.insert(pair<BWAPI::UnitType, UnitCommandQueued>(newCommand.type, newCommand));
    completeCommands.push_back(newCommand);
    completeCommands.sort(SortCompletionTimeLowToHigh);
    return true;

}

void PlayerBuildState::AddCompletedUnit(const UnitCommandQueued& command)
{
    if (command.type == BWAPI::UnitTypes::Terran_SCV && command.unitTask == 0)
    {
        availbleSCVTracker.insert(command.unitId);
        mineralSCVcount++;
    }
    else if (command.type == BWAPI::UnitTypes::Terran_SCV && command.unitTask == 1)
    {
        gasSCVcount++;
    }
    else if (command.type == BWAPI::UnitTypes::Terran_Command_Center)
    {
        baseCount++;
    }
    else if (command.type == BWAPI::UnitTypes::Terran_Refinery)
    {
        refineryCount++;
    }

    completeUnitTypes.insert(pair<BWAPI::UnitType, UnitCommandQueued>(command.type, command));
    completeCommands.push_back(command);
    completeCommands.sort(SortCompletionTimeLowToHigh);

}

const list<UnitCommandQueued>& PlayerBuildState::GetCompleteCommands()
{
    return completeCommands;
}

//pair<int,double> PlayerBuildState::FindLatestBuilderQueue(const BWAPI::UnitType& unitToBuild) //return is builder ID and lastUnitOut frame
//{
//    BWAPI::UnitType builder = unitToBuild.whatBuilds();
//    pair<int, double> soonestQueue = pair<int,double>(-1,-DBL_MAX);
//    //multimap<BWAPI::UnitType,int> completeUnitTypes;
//    for(multimap<BWAPI::UnitType,int>::const_iterator bqueue = completeUnitTypes.lower_bound(builder); bqueue != completeUnitTypes.upper_bound(builder); bqueue++)
//    {
//                
//    }
//    
//}