/* 
 * File:   PlayerBuildState.h
 * Author: chris
 *
 * Created on April 21, 2014, 4:35 PM
 */

#ifndef PLAYERBUILDSTATE_H
#define	PLAYERBUILDSTATE_H

#include <list>
#include "UnitCommandStructs.h"
using namespace std;

class PlayerBuildState
{
public:

    PlayerBuildState() : currentGas(0), currentMinerals(50), mineralSCVcount(5), gasSCVcount(0), baseCount(1), refineryCount(0),
    incompleteCommands(list<UnitCommandQueued>()), completeCommands(list<UnitCommandQueued>()), incompleteUnitTypes(multimap<BWAPI::UnitType, UnitCommandQueued>()),
    completeUnitTypes(multimap<BWAPI::UnitType, UnitCommandQueued>()), buildQueueTracker(multimap<int, pair<int, int> > ()), availbleSCVTracker(set<int>()), buildingSCVTracker(map<int, int>()),
    mineralsPerFramePerSCV(1.0 / 24.0), gasPerFramePerSCV(1.8 / 24.0), scvBuildDelay(75.0), avgMineralNodesPerBase(8.0), avgGasNodesPerBase(2.0),
    scvCapPerMineralNode(3.0f), scvCapPerGasNode(3.0f)

    {
        //        for (int i = 0; i < mineralSCVcount; i++)
        //        {
        //
        //        }
        //        
        //        for (int i = 0; i < gasSCVcount; i++)
        //        {
        //
        //        }
        //        
        //        for (int i = 0; i < baseCount; i++)
        //        {
        //
        //        }
        //        
        //        for (int i = 0; i < refineryCount; i++)
        //        {
        //
        //        }
    }

    PlayerBuildState(const PlayerBuildState& orig) : currentGas(orig.currentGas), currentMinerals(orig.currentMinerals), mineralSCVcount(orig.mineralSCVcount), gasSCVcount(orig.gasSCVcount), baseCount(orig.baseCount), refineryCount(orig.refineryCount),
    incompleteCommands(orig.incompleteCommands), completeCommands(orig.completeCommands), incompleteUnitTypes(orig.incompleteUnitTypes),
    completeUnitTypes(orig.completeUnitTypes), buildQueueTracker(orig.buildQueueTracker), availbleSCVTracker(orig.availbleSCVTracker), buildingSCVTracker(orig.buildingSCVTracker),
    mineralsPerFramePerSCV(orig.mineralsPerFramePerSCV),gasPerFramePerSCV(orig.gasPerFramePerSCV),scvBuildDelay(orig.scvBuildDelay), avgMineralNodesPerBase(orig.avgMineralNodesPerBase), avgGasNodesPerBase(orig.avgGasNodesPerBase), 
    scvCapPerMineralNode(orig.avgGasNodesPerBase), scvCapPerGasNode(orig.scvCapPerGasNode)
    {
    };
    virtual ~PlayerBuildState();

    static void initSupplInfo();

    const bool IncreaseCurrentFrame(const int& frameCount, const int& currentFrame);
    const bool QueueCommand(const UnitCommandRequest& command, const int& currentFrame);
    const bool CheckForCompletePrereqStructs(const UnitCommandRequest& command);
    const bool CheckForAvailableBuildQueue(const UnitCommandRequest& command);
    const bool CheckForRequiredResourcesCost(const UnitCommandRequest& command);
    const bool CheckForRequiredResourcesGathering(const UnitCommandRequest& command);
    const bool NeverSuccessfullyQueued(const UnitCommandRequest& command);
    const bool AddDefaultUnit(const UnitCommandRequest& command);
    const double GetCurrentGas();
    const double GetCurrentMinerals();
    const double GetMineralSCVcount();
    const double GetGasSCVCount();
    const double GetBaseCount();
    const double GetRefineryCount();
    const list<UnitCommandQueued>& GetCompleteCommands();
private:
    static map<BWAPI::UnitType, UnitSupplBuildStats> buildInfo;

    double mineralsPerFramePerSCV;
    double gasPerFramePerSCV;
    double avgMineralNodesPerBase;
    double avgGasNodesPerBase;
    double scvCapPerMineralNode;
    double scvCapPerGasNode;
    double scvBuildDelay;

    double currentGas;
    double currentMinerals;
    double mineralSCVcount;
    double gasSCVcount;
    double baseCount;
    double refineryCount;

    list<UnitCommandQueued> incompleteCommands;
    list<UnitCommandQueued> completeCommands;
    multimap<BWAPI::UnitType, UnitCommandQueued> incompleteUnitTypes;
    multimap<BWAPI::UnitType, UnitCommandQueued> completeUnitTypes;
    //map<int,UnitCommandQueued> unitCommandStatus;
    //builder unit id, being built id, completion frame
    multimap<int, pair<int, int> > buildQueueTracker;
    //special queue trackers only for SCVs, since mineral/gas SCVs are lumped together by BWAPI::UnitType, <scvID, constructing id>
    set<int> availbleSCVTracker;
    map<int, int> buildingSCVTracker;
    //map<BWAPI::UnitType, map<int, list<UnitCommandQueued> > > buildQueueTracker;

    void UpdateUnits(const int& currentFrame);
    void UpdateResources(const int& frameCount);
    const int UnitTypeCompleteCount(const BWAPI::UnitType& unit, const int& task);
    const int UnitTypeIncompleteCount(const BWAPI::UnitType& unit, const int& task);
    //const bool CheckForIncompletePrereqStructs(const UnitCommandRequest& command);
    double ResourcesPerFrame(const int& unitCount, const double& unitRPS, const double& scvCapPerNode, const double& nodes) const;
    pair<int, int> FindEarliestBuilderQueue(const BWAPI::UnitType& unitToBuild); //return is builder ID and lastUnitOut frame
    void AddCompletedUnit(const UnitCommandQueued& command);
    //pair<int,double> FindLatestBuilderQueue(const BWAPI::UnitType& unitToBuild); //return is builder ID and lastUnitOut frame

    //list<CommandCompletion> scvCommands;
    //map<BuildUnitType, map<int, list<CommandCompletion> > > structureCommands; //unit type > unit id > list of commands assigned to it
    //map<int, CommandCompletion> unitCompletionTimes;

};



#endif	/* PLAYERBUILDSTATE_H */

