/* 
 * File:   main.cpp
 * Author: ecsl
 *
 * Created on April 8, 2014, 5:18 PM
 */

#include <cstdlib>
#include <iostream>
#include <cstdlib>
#include <time.h>
#include "BuildEstimator.h"
#include "PlayerBuildState.h"
#include <fstream>
#include <sstream>
#include <string.h>
#include "SparCraft.h"
#include "DEBUG.h"
#include "Game.h"

using namespace std;

enum EncodedBuildCommand
{
    MineralSCV, GasSCV, Marine, Firebat, Vulture, Barracks, Factory, Refinery, Academy, Attack, COMMAND_COUNT
};
void BOSSTest();
 SparCraft::PlayerPtr getSamplePlayer(const IDType playerID);
pair<BWAPI::UnitType, int> DecodeBuildCommand(EncodedBuildCommand command);

int main(int argc, char** argv)
{
    srand(time(NULL));
    SparCraft::init();
    PlayerBuildState::initSupplInfo();
    BOSSTest();
    return 0;
}

void BOSSTest()
{
    BuildEstimator est;
    SparCraft::GameState state;
    SparCraft::PlayerPtr p1 = getSamplePlayer(SparCraft::Players::Player_One);
    SparCraft::PlayerPtr p2 = getSamplePlayer(SparCraft::Players::Player_Two);
    est.AddNewPlayer();
    est.AddNewPlayer();
    vector<bool> attackingPlayer(2, false);

    bool deadlockBlock = true;
    list<EncodedBuildCommand> p1Commands;
    list<EncodedBuildCommand> p2Commands;
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(Barracks);
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Barracks);
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(MineralSCV);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p1Commands.push_back(Marine);
    p2Commands = p1Commands;
    p2Commands.pop_back();
    p2Commands.pop_back();
    p2Commands.pop_back();

    vector<list<EncodedBuildCommand> > playerCommands;
    playerCommands.push_back(p1Commands);
    playerCommands.push_back(p2Commands);


    list<EncodedBuildCommand>::const_iterator p1Command = playerCommands[0].begin();
    list<EncodedBuildCommand>::const_iterator p1End = playerCommands[0].end();
    list<EncodedBuildCommand>::const_iterator p2Command = playerCommands[1].begin();
    list<EncodedBuildCommand>::const_iterator p2End = playerCommands[1].end();
    vector<list<EncodedBuildCommand>::const_iterator> playerBuildOrders;
    vector<list<EncodedBuildCommand>::const_iterator> playerEndBuildOrders;
    playerBuildOrders.push_back(p1Command);
    playerBuildOrders.push_back(p2Command);
    playerEndBuildOrders.push_back(p1End);
    playerEndBuildOrders.push_back(p2End);

    bool attack = false;
    bool remainingCommands = true;
    //bool noCommands = false;
    int frameLimit = 10000; //100,000 frames = 70mins, 10,000frames = 7mins
    while (frameLimit > 0 && !attack && remainingCommands)
    {
        QueueResult nextFrame = QUEUE_RESULT_TYPES;
        remainingCommands = false;
        //noCommands = true;
        for (int player = 0; player < est.GetPlayerCount(); player++)
        {
            DOUT("Player " << player << "'s turn. Frame: " << est.GetCurrentFrame() << endl);
            if (playerBuildOrders[player] != playerEndBuildOrders[player])
            {
                remainingCommands = true;
                DOUT("Action: " << (*playerBuildOrders[player]) << endl);
                if ((*playerBuildOrders[player]) == Attack)
                {
                    attack = true; //end build est loop if a player is ready to attack
                    attackingPlayer[player] = true;
                }
                else
                {
                    pair<BWAPI::UnitType, int> decodedCommand = DecodeBuildCommand(*playerBuildOrders.at(player));
                    QueueResult queueStatus = est.QueueCommand(player, decodedCommand.first, decodedCommand.second);
                    DOUT("Queue Result: " << queueStatus << endl);
                    switch (queueStatus)
                    {
                    case Success:
                        playerBuildOrders[player]++;
                        nextFrame = Success; //succes, try next command before proceeding
                        break;
                    case Stall:
                        //action is valid, but need to move frame forward to issue command(but only after all players need a stall)
                        nextFrame = (nextFrame == QUEUE_RESULT_TYPES ? Stall : nextFrame);
                        break;
                    case Deadlock:
                        if (deadlockBlock) //keep the invalid command, blocking the player from making any more progress
                        {
                            nextFrame = (nextFrame == QUEUE_RESULT_TYPES ? Stall : nextFrame); //set to stall so game can still time out
                        }
                        else
                        {
                            playerBuildOrders[player]++; //ignore invalid command and set frame status to try again before proceeding
                            nextFrame = Deadlock;
                        }
                        break;
                    default:
                        //error
                        break;
                    }
                }
            }
        }
        //if all players want to stall, move onto next frame
        if (nextFrame == Stall)
        {
            est.IncreaseCurrentFrame();
            frameLimit--;
        }
    }

    //BOSS's job complete, use SparCraft to evaluate strength of the build-order from this point forward.
    //prepair combat states
    for (int player = 0; player < est.GetPlayerCount(); player++)
    {
        const list<UnitCommandQueued> battleReadyUnits = est.GetCompleteCommands(player);

        DOUT("\nStarting Lineup for player " << player << ": \n");
        for (list<UnitCommandQueued>::const_iterator unit = battleReadyUnits.begin(); unit != battleReadyUnits.end(); unit++)
        {
            DOUT(unit->type.getName() << " | ");
            if ((!attackingPlayer[player] || !unit->type.isWorker()) && !unit->type.isBuilding())
            {
                DOUT(unit->type.getName() << " | ");
                state.addUnit(unit->type, player, SparCraft::Position(200 + (200 * player), 200 + (200 * player)));
            }
        }
    }

    DOUT(endl);

    int moveLimit = 1000;

    // contruct the game
    SparCraft::Game g(state, p1, p2, moveLimit);

    // play the game
    g.play();

    // you can access the resulting game state after g has been played via getState
    SparCraft::GameState finalState = g.getState();

    // you can now evaluate the state however you wish. let's use an LTD2 evaluation from the point of view of player one
    SparCraft::StateEvalScore p1score = finalState.eval(SparCraft::Players::Player_One, SparCraft::EvaluationMethods::LTD2);
    SparCraft::StateEvalScore p2score = finalState.eval(SparCraft::Players::Player_Two, SparCraft::EvaluationMethods::LTD2);

    DRUN(finalState.print(1));

}

pair<BWAPI::UnitType, int> DecodeBuildCommand(EncodedBuildCommand command)
{
    pair<BWAPI::UnitType, int> newCommand(BWAPI::UnitTypes::None, 0);
    switch (command)
    {
    case MineralSCV:
        newCommand.first = BWAPI::UnitTypes::Terran_SCV;
        break;
    case GasSCV:
        newCommand.first = BWAPI::UnitTypes::Terran_SCV;
        newCommand.second = 1;
        break;
    case Marine:
        newCommand.first = BWAPI::UnitTypes::Terran_Marine;
        break;
    case Firebat:
        newCommand.first = BWAPI::UnitTypes::Terran_Firebat;
        break;
    case Vulture:
        newCommand.first = BWAPI::UnitTypes::Terran_Vulture;
        break;
    case Barracks:
        newCommand.first = BWAPI::UnitTypes::Terran_Barracks;
        break;
    case Factory:
        newCommand.first = BWAPI::UnitTypes::Terran_Factory;
        break;
    case Refinery:
        newCommand.first = BWAPI::UnitTypes::Terran_Refinery;
        break;
    case Academy:
        newCommand.first = BWAPI::UnitTypes::Terran_Academy;
        break;
    default:
        //error
        //return BWAPI::UnitTypes::None;
        break;
    }
    return newCommand;
}

 SparCraft::PlayerPtr getSamplePlayer(const IDType playerID)
{
    // Player is the base class for all Player objects
    //
    // Scripted players all have the same constructor which is just the player ID which will be using this script
    // It is imoprtant to set that player ID correctly, as that player will only be generating and returning moves for that player

     SparCraft::PlayerPtr attackClosest(new  SparCraft::Player_AttackClosest(playerID));

    return attackClosest;
}